#![warn(clippy::pedantic)]

use std::collections::HashSet;
use std::net::{IpAddr, Ipv4Addr, Ipv6Addr};
use std::path::{Path, PathBuf};

use crate::config::{Config, RecordType};

use anyhow::{Context, Result};
use clap::{Parser, Subcommand};
use reqwest::Url;
use serde::{Deserialize, Serialize};
use serde_json::json;
use tabled::settings::object::Column;
use tabled::settings::{Alignment, Modify};
use tabled::{Table, Tabled};
use tracing::{debug, error, info, warn};

mod config;

const X_AUTH_EMAIL: &str = "X-Auth-Email";
const X_AUTH_KEY: &str = "X-Auth-Key";

#[derive(Parser, Clone, Debug)]
pub struct Args {
    /// Path to the configuration file.
    ///
    /// If not provided, checks the current working directory, the current
    /// user's local config directory, and finally the system wide config
    /// directory.
    #[clap(short, long, global = true)]
    config_file: Option<PathBuf>,
    #[command(subcommand)]
    cmd: Command,
}

#[derive(Subcommand, Clone, Debug)]
pub enum Command {
    Run,
    List(List),
}

#[derive(Parser, Clone, Debug)]
pub struct List {
    zones: Option<Vec<String>>,
}

#[tokio::main]
async fn main() -> Result<()> {
    let args = Args::parse();
    let config = load_config(args.config_file).context("Failed to find a suitable config file")?;
    match args.cmd {
        Command::Run => handle_run(config).await,
        Command::List(list) => handle_list(config, list).await,
    }
}

async fn handle_run(conf: Config) -> Result<()> {
    let ipv4 = match conf.ip_reflector.ipv4 {
        Some(addr_to_req) => Some(
            get_ipv4(addr_to_req)
                .await
                .context("Failed to query for ipv4 address, bailing.")?,
        ),
        None => None,
    };
    let ipv6 = match conf.ip_reflector.ipv6 {
        Some(addr_to_req) => Some(
            get_ipv6(addr_to_req)
                .await
                .context("Failed to query for ipv4 address, bailing.")?,
        ),
        None => None,
    };

    let ip_cache_path = ip_cache_path().context("while getting the ip cache path")?;

    // Only start processing requests if our current IP doesn't match our cached IP
    if !needs_update(&ip_cache_path, ipv4, ipv6).context("while checking the ip cache")? {
        return Ok(());
    }

    let ipv4_addr = ipv4.map(IpAddr::V4);
    let ipv6_addr = ipv6.map(IpAddr::V6);

    for zone in conf.zone.into_values() {
        let zone_id = zone.id;

        let records_to_process = zone.record.into_iter().filter_map(|record| {
            if record.disabled {
                return None;
            }

            if ipv4.is_some() && record.is_ipv4() {
                return Some((&ipv4_addr, record));
            }

            if ipv6.is_some() && record.is_ipv6() {
                return Some((&ipv6_addr, record));
            }

            None
        });

        for (addr, record) in records_to_process {
            #[derive(Deserialize, Debug)]
            #[allow(dead_code)]
            struct UpdateDnsResponse {
                success: bool,
                errors: Vec<Message>,
                messages: Vec<Message>,
            }

            let record_id = record.id;
            let resp: UpdateDnsResponse = reqwest::Client::new()
                .put(format!(
                    "https://api.cloudflare.com/client/v4/zones/{zone_id}/dns_records/{record_id}"
                ))
                .header(X_AUTH_EMAIL, &conf.account.email.to_string())
                .header(X_AUTH_KEY, &conf.account.api_key)
                .json(&json!({
                    "type": record.protocol_type,
                    "name": record.name,
                    "content": addr,
                    "ttl": 1, // Auto TTL
                    "proxied": record.proxy,
                }))
                .send()
                .await
                .context("while requesting an api endpoint")?
                .json()
                .await
                .context("while parsing into a json")?;

            if resp.success {
                continue;
            }
        }
    }

    // Updating the ip cache last is better in case we get interrupted. Better
    // to update too frequently than not enough.
    update_ip_cache(ip_cache_path, ipv4, ipv6)?;
    Ok(())
}

fn update_ip_cache<P: AsRef<Path>>(
    path: P,
    ipv4: Option<Ipv4Addr>,
    ipv6: Option<Ipv6Addr>,
) -> Result<()> {
    let data = serde_json::to_string(&CacheFile { ipv4, ipv6 }).expect("serialization to work");
    std::fs::write(path, data).context("while updating the ip cache file")?;
    Ok(())
}

#[derive(Serialize, Deserialize, Default)]
struct CacheFile {
    ipv4: Option<Ipv4Addr>,
    ipv6: Option<Ipv6Addr>,
}

fn needs_update<P: AsRef<Path>>(
    path: P,
    cur_ipv4: Option<Ipv4Addr>,
    cur_ipv6: Option<Ipv6Addr>,
) -> Result<bool> {
    let data = std::fs::read_to_string(path).context("while reading the ip cache file")?;
    let cache: CacheFile = match serde_json::from_str(&data) {
        Ok(cache) => cache,
        Err(e) => {
            warn!("Failed to parse the ip cache file; assuming empty: {e}");
            CacheFile::default()
        }
    };

    if matches!((cache.ipv4, cur_ipv4), (Some(_), None)) {
        warn!("Cache previously reported an IPv4 address, but we didn't find one. Not updating IPv4 entries");
    }

    let update_ipv4 = match (cache.ipv4, cur_ipv4) {
        (Some(cached), Some(current)) => cached != current,
        (cached, current) => cached.xor(current).is_some(),
    };

    if matches!((cache.ipv6, cur_ipv6), (Some(_), None)) {
        warn!("Cache previously reported an IPv6 address, but we didn't find one. Not updating IPv6 entries");
    }

    let update_ipv6 = match (cache.ipv6, cur_ipv6) {
        (Some(cached), Some(current)) => cached != current,
        (cached, current) => cached.xor(current).is_some(),
    };

    Ok(update_ipv4 || update_ipv6)
}

fn ip_cache_path() -> Result<PathBuf> {
    dirs::cache_dir()
        .context("Failed to determine cache directory")
        .map(|path| path.join("cloudflare-ddns.cache"))
}

#[derive(Deserialize, Debug)]
#[allow(dead_code)]
struct Message {
    code: u16,
    message: String,
}

async fn handle_list(conf: Config, args: List) -> Result<()> {
    // Use provided zones or list all in config
    let known_zones: HashSet<_> = conf.zone.values().map(|zone| &zone.id).collect();
    let zones: Vec<_> = match args.zones {
        Some(zones) => {
            // These zones may be human readable. Map them to zone IDs.
            zones
                .into_iter()
                .filter_map(|maybe_zone_id| {
                    if known_zones.contains(&maybe_zone_id) {
                        return Some(maybe_zone_id);
                    }
                    if let Some(zone) = conf.zone.get(&maybe_zone_id) {
                        return Some(zone.id.clone());
                    }

                    eprintln!("Unknown zone {maybe_zone_id}, skipping");
                    None
                })
                .collect()
        }
        None => known_zones.into_iter().cloned().collect(),
    };

    for zone in zones {
        #[derive(Deserialize, Debug)]
        #[allow(dead_code)]
        struct ListZoneResponse {
            success: bool,
            errors: Vec<Message>,
            messages: Vec<Message>,
            result: Vec<DnsResponse>,
        }

        #[derive(Deserialize, Debug, Tabled)]
        #[tabled(rename_all = "PascalCase")]
        struct DnsResponse {
            name: String,
            #[tabled(rename = "Type")]
            r#type: RecordType,
            #[tabled(rename = "IP Address")]
            content: IpAddr,
            proxied: bool,
            id: String,
        }

        let mut entries = vec![];
        for page_no in 1.. {
            // This technically requests one more than optimal, but tbh it
            // doesn't really matter
            let resp: ListZoneResponse = reqwest::Client::new()
                .get(format!(
                "https://api.cloudflare.com/client/v4/zones/{zone}/dns_records?type=A,AAAA&page={page_no}"
            ))
                .header(X_AUTH_EMAIL, &conf.account.email.to_string())
                .header(X_AUTH_KEY, &conf.account.api_key)
                .send()
                .await
                .context("while requesting an api endpoint")?
                .json()
                .await
                .context("while parsing into a json")?;

            // todo: handle messages, errors, and non-success response

            if resp.result.is_empty() {
                break;
            } else {
                entries.extend(resp.result);
            }
        }

        // Sort by subdomain, with higher level subdomains taking higher precedence than lower ones.
        entries.sort_unstable_by(|a, b| a.name.split('.').rev().cmp(b.name.split('.').rev()));

        println!(
            "{}",
            Table::new(entries).with(Modify::new(Column::from(0)).with(Alignment::right()))
        );
    }

    Ok(())
}

async fn get_ipv4(url: Url) -> Result<Ipv4Addr> {
    reqwest::get(url)
        .await
        .context("Failed send IPv4 reflector request")?
        .text()
        .await
        .context("Failed to get IPv4 reflector data")?
        .parse()
        .context("Response was not an IPv4 address")
}

async fn get_ipv6(url: Url) -> Result<Ipv6Addr> {
    reqwest::get(url)
        .await
        .context("Failed send IPv4 reflector request")?
        .text()
        .await
        .context("Failed to get IPv4 reflector data")?
        .parse()
        .context("Response was not an IPv4 address")
}

fn load_config(user_provided_path: Option<PathBuf>) -> Option<Config> {
    if let Some(path) = user_provided_path {
        return load_config_from_path(path);
    }

    let file_path = Path::new("./cloudflare-ddns.toml");
    let resolved_path = file_path.canonicalize();
    let resolved_path = resolved_path.as_deref().unwrap_or(file_path);
    if let Some(config) = load_config_from_path(resolved_path) {
        return Some(config);
    }

    if let Some(config) = dirs::config_dir()
        .map(|path| path.join(file_path))
        .and_then(load_config_from_path)
    {
        return Some(config);
    }

    load_config_from_path("/etc/cloudflare-ddns.toml")
}

fn load_config_from_path<P: AsRef<Path>>(path: P) -> Option<Config> {
    match std::fs::read_to_string(&path) {
        Ok(data) => match toml::from_str(&data) {
            Ok(config) => return Some(config),
            Err(err) => {
                debug!(
                    "Failed to parse config file at {}: {}",
                    path.as_ref().to_string_lossy(),
                    err
                )
            }
        },
        Err(err) => {
            debug!(
                "Unable to read the config file at {}: {}",
                path.as_ref().to_string_lossy(),
                err
            )
        }
    }
    None
}
